/*
 * spcplay - main (program entry point)
 *
 * Copyright (C) 2000 AGAWA Koji <kaoru-k@self-core.104.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* $Id: main.c,v 1.1.1.1 2000/06/13 19:25:54 kaoru-k Exp $ */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include "getopt.h"
#include "libspc.h"

static int is_verbose;

static int get_options (int argc, char **argv);
static void usage (void);
static void display_version (void);

SPC_Config spc_config = {
    44100,
    16,
    2,
    1,
    1
};

int main(int argc, char **argv)
{
    int buf_size, num_options;
    void *buf;
  
    num_options = get_options(argc, argv);

    buf_size = SPC_init(&spc_config);
    buf = malloc(buf_size);
    SPC_load(argv[num_options]);
    for (;;) {
	SPC_update(buf);
	fwrite(buf, 1, buf_size, stdout);
      }
    SPC_close();

    return 0;
}

/*
   ---------------------------------------------------------------- */
static int get_options (int argc, char **argv)
{
  // extern char *optarg;
  // extern int optind, opterr, optopt;

  while (1)
    {
      int c;
      int option_index = 0;
      static struct option long_options[] = {
	{"verbose",     0, 0, 'V'},
	{"version",     0, 0, 'v'},
	{"help",        0, 0, 'h'},
	{0, 0, 0, 0}
      };

      c = getopt_long (argc, argv, "Vvh", long_options, &option_index);
      if (c == EOF)
	break;

      switch (c)
	{
	case 'V':
/* 	  is_verbose = TRUE; */
	  break;
	  
	case 'v':
	  display_version ();
	  break;
	  
        case 'h':
	  usage ();
	  break;
	  
	case '?':
	  exit (1);
	  break;
	  
	default:
	  exit (1);
	  break;
	}
    }
  
  return optind;
}

/*
   ---------------------------------------------------------------- */
static void usage (void)
{
  fprintf (stderr, "usage: %s [options]\n", PACKAGE);
  fprintf (stderr, "Options:\n");
  fprintf (stderr, "-V,      --verbose         ");
  fprintf (stderr, "Verbose mode.\n");
  fprintf (stderr, "-v,      --version         ");
  fprintf (stderr, "Show version information.\n");
  fprintf (stderr, "-h,      --help            ");
  fprintf (stderr, "Show this help message.\n");

  exit (0);
}

/*
   ---------------------------------------------------------------- */
static void display_version (void)
{
  fprintf (stderr, "%s version " VERSION "\n", PACKAGE);
  fprintf (stderr, "Copyright (C) 2000 AGAWA Koji <kaoru-k@self-core.104.net>\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "This is free software; see the source for copying conditions.\n");
  fprintf (stderr, "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n");
  fprintf (stderr, "PARTICULAR PURPOSE.\n\n");

  exit (0);
}
