/*
 * SPC input plugin for XMMS - GUI
 *
 * Copyright (C) 2000 AGAWA Koji <kaoru-k@self-core.104.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include "spcxmms.h"
#include "interface.h"
#include "support.h"
#include "libspc.h"

/* Widgets
   ---------------------------------------------------------------- */
GtkWidget *w_configure_window = NULL;
GtkCombo *w_sampling_rate;
GtkToggleButton *w_resolution_16;
GtkToggleButton *w_resolution_8;
GtkToggleButton *w_channels_2;
GtkToggleButton *w_channels_1;
GtkToggleButton *w_interpolation;
GtkToggleButton *w_echo;

static GtkWidget *w_tag = NULL;
static GtkEntry *w_songname;
static GtkEntry *w_gametitle;
static GtkEntry *w_dumpername;
static GtkEntry *w_author;
static GtkEntry *w_comments;
static GtkEntry *w_date_yyyy;
static GtkEntry *w_date_mm;
static GtkEntry *w_date_dd;
static GtkEntry *w_playtime;
static GtkEntry *w_fadetime;
static SPC_ID666 *spc_id666 = NULL;
static gchar *spc_id666_filename = NULL;

gint
option_menu_get_history (GtkWidget *omenu)
{
  GtkWidget *menu, *menu_item;
  GList *children, *child;

  if (omenu) {
    menu = gtk_option_menu_get_menu(GTK_OPTION_MENU(omenu));
    menu_item = GTK_OPTION_MENU(omenu)->menu_item;
    if (menu) {
      children = GTK_MENU_SHELL(menu)->children;
      child = g_list_find(children, menu_item);
      if (child) {
	return g_list_position(children, child);
      }
    }
  }

  return -1;
}

/* ---------------------------------------------------------------- */
static void tag_set_widgets (GtkWidget *w)
{
  w_songname = GTK_ENTRY (lookup_widget (w, "tag_songname"));
  w_gametitle = GTK_ENTRY (lookup_widget (w, "tag_gametitle"));
  w_dumpername = GTK_ENTRY (lookup_widget (w, "tag_dumper"));
  w_author = GTK_ENTRY (lookup_widget (w, "tag_author"));
  w_comments = GTK_ENTRY (lookup_widget (w, "tag_comments"));
  w_date_yyyy = GTK_ENTRY (lookup_widget (w, "tag_date_yyyy"));
  w_date_mm = GTK_ENTRY (lookup_widget (w, "tag_mm"));
  w_date_dd = GTK_ENTRY (lookup_widget (w, "tag_dd"));
  w_playtime = GTK_ENTRY (lookup_widget (w, "tag_playtime"));
  w_fadetime = GTK_ENTRY (lookup_widget (w, "tag_fadetime"));
}

static void tag_set_states (SPC_ID666 *id)
{
  gtk_entry_set_text (w_songname, id->songname);
  gtk_entry_set_text (w_gametitle, id->gametitle);
  gtk_entry_set_text (w_dumpername, id->dumper);
  gtk_entry_set_text (w_author, id->author);
  gtk_entry_set_text (w_comments, id->comments);
}

void file_info_box (char *filename)
{
  if (!w_tag)
    {
      w_tag = create_tag ();
      tag_set_widgets (w_tag);

      spc_id666 = SPC_get_id666 (filename);
      if (spc_id666)
	{
	  tag_set_states (spc_id666);
	  spc_id666_filename = g_strdup (filename);
	}
      
      gtk_widget_show (w_tag);
    }
  else
    gdk_window_raise (w_tag->window);
}

void
on_tag_save_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  if (spc_id666)
    {
      SPC_write_id666 (spc_id666, spc_id666_filename);

      free (spc_id666);
      spc_id666 = NULL;
      free (spc_id666_filename);
      spc_id666_filename = NULL;
    }

  gtk_widget_destroy (w_tag);
  w_tag = NULL;
}


void
on_tag_remove_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{

}

void
on_tag_cancel_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (w_tag);
  w_tag = NULL;

  if (spc_id666)
    {
      free (spc_id666);
      spc_id666 = NULL;
      free (spc_id666_filename);
      spc_id666_filename = NULL;
    }
}

/* configure window
   ================================================================ */
static void configure_set_widgets (GtkWidget *w)
{
  w_sampling_rate = GTK_COMBO (lookup_widget (w, "sampling_rate"));
  w_resolution_16 = GTK_TOGGLE_BUTTON (lookup_widget (w, "resolution_16"));
  w_resolution_8 = GTK_TOGGLE_BUTTON (lookup_widget (w, "resolution_8"));
  w_channels_2 = GTK_TOGGLE_BUTTON (lookup_widget (w, "channels_2"));
  w_channels_1 = GTK_TOGGLE_BUTTON (lookup_widget (w, "channels_1"));
  w_interpolation = GTK_TOGGLE_BUTTON (lookup_widget (w, "interpolation"));
  w_echo = GTK_TOGGLE_BUTTON (lookup_widget (w, "echo"));
}

static void configure_set_states (PluginConfig *cfg)
{
  char buf[16];

  sprintf (buf, "%d", cfg->spc.sampling_rate);
  gtk_entry_set_text (GTK_ENTRY (w_sampling_rate->entry), buf);
			  
  if (cfg->spc.resolution == 16)
    gtk_toggle_button_set_state (w_resolution_16, TRUE);
  else
    gtk_toggle_button_set_state (w_resolution_8, TRUE);

  if (cfg->spc.channels == 2)
    gtk_toggle_button_set_state (w_channels_2, TRUE);
  else
    gtk_toggle_button_set_state (w_channels_1, TRUE);

  gtk_toggle_button_set_state (w_interpolation, cfg->spc.is_interpolation);
  gtk_toggle_button_set_state (w_echo, cfg->spc.is_echo);
}

static void configure_get_states (PluginConfig *cfg)
{
  int rate;

  rate = atoi (gtk_entry_get_text (GTK_ENTRY (w_sampling_rate->entry)));
  if (rate > 1)
    cfg->spc.sampling_rate = rate;

  cfg->spc.resolution = w_resolution_16->active ? 16 : 8;
  cfg->spc.channels = w_channels_2->active ? 2 : 1;
  cfg->spc.is_interpolation = w_interpolation->active;
  cfg->spc.is_echo = w_echo->active;
}

void configure (void)
{
  if (!w_configure_window)
    {
      w_configure_window = create_configure ();
      configure_set_widgets (w_configure_window);
      configure_set_states (&config);
      gtk_widget_show (w_configure_window);
    }
  else
    gdk_window_raise (w_configure_window->window);
}

  
void on_config_changed (void);

void
on_cf_ok_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
  configure_get_states (&config);
  config_write_to_default_file (&config);
  on_config_changed ();

  gtk_widget_destroy (w_configure_window);
  w_configure_window = NULL;
}

void
on_cf_cancel_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (w_configure_window);
  w_configure_window = NULL;
}

void
on_cf_apply_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  configure_get_states (&config);
  config_write_to_default_file (&config);
  on_config_changed ();
}

